"""
Copyright 2021 Charles McMarrow

This script translates "loads" a program into Rooms "memory".
It verifies that the data that it loads into Rooms is valid.
It lets data to be translated VIA file or str.
"""

# built-in
import io
import os
import string
from collections import deque
from typing import Iterator, Optional, Tuple, Dict, List

# backrooms
from .backrooms_error import BackroomsError
from .rooms import Rooms

INCLUDE_FILE_EXTENSION = ".brs"


class TranslatorError(BackroomsError):
    @classmethod
    def missing_include(cls,
                        include: str) -> 'TranslatorError':
        """
        info: Used to indicate a missing include.
        :param include: str
        :return: TranslatorError
        """
        return cls(f"{repr(include)} is a missing include!")

    @classmethod
    def name_collision(cls, name: str):
        return cls(f"Name collision with {name}!")

    @classmethod
    def bad_line(cls,
                 line: str,
                 line_number: int,
                 handler_name: str,
                 error: str = "") -> 'TranslatorError':
        """
        info: Used to indicate line is not valid.
        :param line: str
        :param line_number: int
        :param handler_name: str
        :param error: str
        :return: TranslatorError
        """
        if not error:
            return cls(f"Bad line at {handler_name}:{line_number}: {repr(line)}!")
        return cls(f"Bad line at {handler_name}:{line_number}:{str(error)}: {repr(line)}!")


class Handler:
    def __init__(self,
                 name: str):
        """
        info: Gives data line by line.
        :param name: str
        """
        self._name: str = name
        self._is_open = True

    def get_name(self) -> str:
        """
        info: Gets the name of the Handler.
        :return: str
        """
        return self._name

    def is_open(self) -> bool:
        """
        info: Checks if file handler is still open.
        :return: bool
        """
        return self._is_open

    def close(self) -> None:
        """
        info: Closes the handler.
        :return: None
        """
        self._is_open = False

    def __del__(self):
        self.close()
    
    def __iter__(self) -> 'Handler':
        return self
    
    def __next__(self) -> str:
        raise NotImplementedError()


class FileHandler(Handler):
    def __init__(self,
                 path: str):
        """
        info: Handler for Files.
        :param path: str
        """
        self._path: str = path
        self._file_iter: Optional[Iterator[str]] = None
        self._file_handler: Optional[io.TextIOWrapper] = None
        file_name = os.path.basename(path)
        if file_name.endswith(INCLUDE_FILE_EXTENSION):
            file_name = file_name[:-len(INCLUDE_FILE_EXTENSION)]
        super(FileHandler, self).__init__(file_name)

    def close(self):
        if self.is_open() and self._file_handler is not None:
            self._file_handler.close()
        super(FileHandler, self).close()

    def __next__(self) -> str:
        if self._file_iter is None:
            self._file_handler = open(self._path)
            self._file_iter = iter(self._file_handler)
        return next(self._file_iter).rstrip("\n")


class StringHandler(Handler):
    def __init__(self,
                 name: str,
                 data: str):
        """
        info: Handler for Strings
        :param name: str
        :param data: str
        """
        self._parts: Iterator[str] = iter(data.split("\n"))
        super(StringHandler, self).__init__(name)

    def __next__(self) -> str:
        return next(self._parts)
        

class Handlers:
    def __init__(self,
                 main: Handler,
                 name_spaces: Tuple[Tuple[Handler, ...], ...] = ()):
        """
        info: Holder all the Handles need it for a program.
            Builds the name_spaces so includes can be prioritized.
            Makes sure a handler is not included more than once.
        :param main: Handler
        :param name_spaces: Tuple[Tuple[Handler]])
        """
        self._handlers: deque[Handler] = deque((main,))
        self._name_spaces: Optional[Tuple[Dict[str, Handler]]] = None

        name_spaces_dicts = []
        for name_space in name_spaces:
            name_space_dict = {}
            for handler in name_space:
                if handler.get_name() in name_space_dict:
                    raise TranslatorError.name_collision(handler.get_name())
                name_space_dict[handler.get_name()] = handler
            name_spaces_dicts.append(name_space_dict)
        self._name_spaces: Tuple[Dict[str, Handler]] = tuple(name_spaces_dicts)

        self._line_number: int = -1
        self._used_names = {main.get_name()}
    
    def __iter__(self) -> 'Handlers':
        return self
    
    def __next__(self) -> str:
        while self._handlers:
            try:
                self._line_number += 1
                return next(self._handlers[0])
            except StopIteration:
                if self._line_number == -1:
                    # return a single bank line from the handler with no data
                    self._line_number = 0
                    return ""
                self._handlers.popleft()
                self._line_number = -1
        raise StopIteration()

    def __bool__(self):
        return bool(self._handlers)

    def get_name(self) -> Optional[str]:
        """
        info: Gets the current name.
        :return: Optional[str]
        """
        if self._handlers:
            return self._handlers[0].get_name()

    def get_line_number(self) -> Optional[int]:
        """
        info: Gets the current line.
        :return: Optional[int]
        """
        if self._line_number >= 0:
            return self._line_number

    def include(self,
                name: str) -> None:
        """
        info: Includes name if it has not been included all ready.
        :param name: str
        :exception TranslatorError
            raises TranslatorError is include cant be found.
        :return: None
        """
        # check that name has not be included yet
        if name not in self._used_names:
            for name_space in self._name_spaces:
                if name in name_space:
                    self._used_names.add(name)
                    self._handlers.append(name_space[name])
                    return
            # could not find the include handler
            raise TranslatorError.missing_include(name)


def _tokenize_line(line: str) -> List[str]:
    """
    info: Tokenize line.
    :param line: str
    :return: List[str]
    """
    tokens = [""]
    for character in line:
        # split and remove white space
        if character in string.whitespace:
            if tokens[-1]:
                tokens.append("")
        # split and add "@" which acts as None
        elif character == "@":
            if not tokens[-1]:
                tokens[-1] += "@"
            else:
                tokens.append("@")
            tokens.append("")
        else:
            tokens[-1] += character
    if not tokens[-1]:
        del tokens[-1]
    return tokens


def _read_number(tokens: List[str],
                 full_line: str,
                 line_number: int,
                 handler_name: str) -> int:
    """
    info: Checks and pops an int from tokens.
    :param tokens: List[str]
    :param full_line: str
    :param line_number: int
    :param handler_name: str
    :exception TranslatorError
            raises TranslatorError no tokens was given.
    :exception ValueError
            raises ValueError if int() fails.
    :return: int
    """
    if tokens:
        return int(tokens.pop(0))
    raise TranslatorError.bad_line(full_line, line_number, handler_name)


def _read_single_number(line: str,
                        full_line: str,
                        line_number: int,
                        handler_name: str) -> int:
    """
    info: Checks that line holds only a single int.
    :param line: str
    :param full_line: str
    :param line_number: int
    :param handler_name: str
    :exception TranslatorError
            raises TranslatorError if line makes more or less than one token.
    :exception ValueError
            raises ValueError if int() fails.
    :return: int
    """
    tokens = _tokenize_line(line)
    if len(tokens) == 1:
        return int(tokens[0])
    raise TranslatorError.bad_line(full_line, line_number, handler_name)


def translator(handlers: Handlers) -> Rooms:
    """
    info: Load program into Rooms "memory"
    :param handlers: Handlers
    :exception TranslatorError
            raises TranslatorError if handlers give an invalid line.
    :return: Rooms
    """
    rooms = Rooms()
    x, y, floor = 0, 0, 1
    for line in handlers:
        full_line = line
        # throw away lead white space
        while line:
            if line[0] not in string.whitespace:
                break
            line = line[1:]
        try:
            # enter new handler
            if handlers.get_line_number() == 0:
                # reset x and y to zero
                x = 0
                y = 0
                # go down a floor
                floor += -1
                # set the floor name to the handlers name
                rooms.set_floor_name(floor, handlers.get_name())

            # row
            if line.startswith("/"):
                # write row into rooms
                rooms.write_line(x, y, floor, line[1:], 1, 0, 0)
                # go down a row
                y += -1
            # comment
            elif line.startswith("#"):
                # just throw away the comment
                pass
            # hallway
            elif line.startswith("~"):
                tokens = _tokenize_line(line[1:])
                hallway_name = None
                # check if hallway has a name
                if tokens:
                    if tokens[0] == "@":
                        del tokens[0]
                    else:
                        hallway_name = tokens.pop(0)
                    # check that we are out of tokens
                    if tokens:
                        raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
                rooms.set_hallway_name(y, floor, hallway_name)
            # new floor
            elif line.startswith("+"):
                tokens = _tokenize_line(line[1:])
                floor_name = None
                # check if floor has a name
                if tokens:
                    if tokens[0] == "@":
                        del tokens[0]
                    else:
                        floor_name = tokens.pop(0)
                    # check that we are out of tokens
                    if tokens:
                        raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
                floor += -1
                y = 0
                rooms.set_floor_name(floor, floor_name)
            # include
            elif line.startswith("%"):
                tokens = _tokenize_line(line[1:])
                # check that we have include name
                if tokens:
                    handlers.include(tokens.pop(0))
                    # check that we are out of tokens
                    if tokens:
                        raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
                else:
                    raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
            # parallel
            elif line.startswith("="):
                tokens = _tokenize_line(line[1:])

                from_name = None
                to_name = None
                from_spot = None
                to_spot = None

                if tokens:
                    if tokens[0] != "@":
                        from_name = tokens.pop(0)
                    else:
                        tokens.pop(0)
                    if tokens:
                        if tokens[0] != "@":
                            to_name = tokens.pop(0)
                        else:
                            tokens.pop(0)
                        if tokens:
                            if tokens[0] != "@":
                                from_spot = int(tokens.pop(0))
                            else:
                                tokens.pop(0)
                            if tokens:
                                if tokens[0] != "@":
                                    to_spot = int(tokens.pop(0))
                                else:
                                    tokens.pop(0)
                                if tokens:
                                    raise TranslatorError.bad_line(full_line,
                                                                   handlers.get_line_number(),
                                                                   handlers.get_name())

                if from_name is not None and from_spot is not None:
                    raise TranslatorError.bad_line(full_line,
                                                   handlers.get_line_number(),
                                                   handlers.get_name(),
                                                   "Cant provided both 'from_name' and 'from_location'!")

                if from_name is not None:
                    # duplicate from name
                    from_spot = rooms.get_floor_level(from_name)
                elif from_spot is None:
                    from_spot = floor

                if to_spot is None:
                    # duplicate onto the floor above
                    to_spot = floor - 1
                    floor += -1

                # duplicate floor
                rooms.duplicate_floor(from_spot, to_spot)
                # set floor name
                rooms.set_floor_name(to_spot, to_name)
            # shift x
            elif line.startswith("XS"):
                x += _read_single_number(line[2:], full_line, handlers.get_line_number(), handlers.get_name())
            # x
            elif line.startswith("X"):
                x = _read_single_number(line[1:], full_line, handlers.get_line_number(), handlers.get_name())
            # shift y
            elif line.startswith("YS"):
                y += _read_single_number(line[2:], full_line, handlers.get_line_number(), handlers.get_name())
            # y
            elif line.startswith("Y"):
                y = _read_single_number(line[1:], full_line, handlers.get_line_number(), handlers.get_name())
            # floor shift
            elif line.startswith("FS"):
                floor += _read_single_number(line[2:], full_line, handlers.get_line_number(), handlers.get_name())
            # floor
            elif line.startswith("F"):
                floor = _read_single_number(line[1:], full_line, handlers.get_line_number(), handlers.get_name())
            # blank line
            elif not line:
                # just throw away the blank line
                pass
            # unknown line
            else:
                raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
        except TranslatorError as e:
            raise e
        except Exception as e:
            raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name(), str(e))
    return rooms


def load_dir(file: str) -> Tuple[Handler, Tuple[Handler, ...]]:
    """
    info: Loads all files in a file dir into handlers.
    :param file: str
    :return: Tuple[Handler, Tuple[Handler, ...]]
    """
    # TODO test
    file = os.path.abspath(file)
    main_handler = FileHandler(file)
    handlers = []
    file_dir = os.path.join(*os.path.split(file)[:-1])
    for file in os.listdir(file_dir):
        file = os.path.join(file_dir, file)
        if os.path.isfile(file):
            handlers.append(FileHandler(file))
    return main_handler, tuple(handlers)



"""
Copyright 2021 Charles McMarrow
"""

# built-in
from typing import Tuple

# backrooms
from backrooms.translator import StringHandler
from . import variables


def get_builtins() -> Tuple[StringHandler, ...]:
    """
    info: Gets Backrooms builtins modules.
    :return: Tuple[StringHandler, ...]
    """
    return variables.get_handler(),



"""
Copyright 2021 Charles McMarrow

This script holds a simple conscious "thread" data structure.
A conscious holds it own state.
    * work stack
    * function ret stack
    * 10 work registers
    * an error register
    * PC registers
"""


# built-in
from typing import Dict, Tuple, Union

# backrooms
from . import stack


# Stacks
WORK_STACK = "WORK_STACK"
FUNCTION_STACK = "FUNCTION_RET_STACK"

# Normal Registers
WORKING_REGISTER = "WORKING_REGISTER"
R0 = "R0"
R1 = "R1"
R2 = "R2"
R3 = "R3"
R4 = "R4"
R5 = "R5"
R6 = "R6"
R7 = "R7"
R8 = "R8"
R9 = "R9"

# Program Counter Registers
PC_X = "PC_X"
PC_Y = "PC_Y"
PC_FLOOR = "PC_FLOOR"

# Program Counter Vector Registers
PC_V_X = "PC_V_X"
PC_V_Y = "PC_V_Y"
PC_V_FLOOR = "PC_V_FLOOR"

# Thread ID
ID = "ID"

# State
ALIVE = "ALIVE"
HALT = "HALT"


# TODO test
def _to_branch_int(obj: Union[int, str, stack.StackFrame, stack.StackBottom, None]) -> int:
    if isinstance(obj, str):
        return int(bool(obj))
    elif obj is None:
        return 0
    elif obj is stack.StackFrame:
        return 0
    elif obj is stack.StackBottom:
        return 0
    return obj


def _clear(conscious: 'Conscious') -> True:
    return True


def _less_than_zero(conscious: 'Conscious') -> True:
    return _to_branch_int(conscious[WORK_STACK].peak()) < 0


def _greater_than_zero(conscious: 'Conscious') -> True:
    return _to_branch_int(conscious[WORK_STACK].peak()) > 0


def _zero(conscious: 'Conscious') -> True:
    return _to_branch_int(conscious[WORK_STACK].peak()) == 0


def _not_zero(conscious: 'Conscious') -> True:
    return _to_branch_int(conscious[WORK_STACK].peak()) != 0


def _is_integer(conscious: 'Conscious') -> True:
    return isinstance(conscious[WORK_STACK].peak(), int)


def _is_string(conscious: 'Conscious') -> True:
    return isinstance(conscious[WORK_STACK].peak(), str)


def _is_none(conscious: 'Conscious') -> True:
    return conscious[WORK_STACK].peak() is None


def _is_stack_frame(conscious: 'Conscious') -> True:
    return conscious[WORK_STACK].peak() is stack.StackFrame


def _is_stack_bottom(conscious: 'Conscious') -> True:
    return conscious[WORK_STACK].peak() is stack.StackBottom


# Branch
BRANCH = "BRANCH"
BRANCH_CLEAR = _clear
BRANCH_LESS_THAN_ZERO = _less_than_zero
BRANCH_GREATER_THAN_ZERO = _greater_than_zero
BRANCH_ZERO = _zero
BRANCH_NOT_ZERO = _not_zero
BRANCH_IS_INTEGER = _is_integer
BRANCH_IS_STRING = _is_string
BRANCH_IS_NONE = _is_none
BRANCH_IS_STACK_FRAME = _is_stack_frame
BRANCH_IS_STACK_BOTTOM = _is_stack_bottom


class Conscious(Dict):
    def __init__(self, **kwargs):
        """
        info: Makes a conscious "Thread".
        :param kwargs: Dict[str, object]
        :return: Dict[str, object]
        """
        new_conscious = {WORK_STACK: stack.Stack(),
                         FUNCTION_STACK: stack.Stack(),
                         WORKING_REGISTER: R0,
                         R0: None,
                         R1: None,
                         R2: None,
                         R3: None,
                         R4: None,
                         R5: None,
                         R6: None,
                         R7: None,
                         R8: None,
                         R9: None,
                         PC_X: 0,
                         PC_Y: 0,
                         PC_FLOOR: 0,
                         PC_V_X: 1,
                         PC_V_Y: 0,
                         PC_V_FLOOR: 0,
                         ID: None,
                         ALIVE: True,
                         HALT: False,
                         BRANCH: BRANCH_CLEAR}
        new_conscious.update(kwargs)
        super(Conscious, self).__init__(new_conscious)

    def step(self) -> None:
        """
        info: Shift the PC Registers based off the PC Vector Registers.
        :return:
        """
        self[PC_X] += self[PC_V_X]
        self[PC_Y] += self[PC_V_Y]
        self[PC_FLOOR] += self[PC_V_FLOOR]

    def next_step(self) -> Tuple[int, int, int]:
        """
        info: Get the next location for conscious.
        :return: Tuple[int, int, int]
        """
        return self[PC_X] + self[PC_V_X], self[PC_Y] + self[PC_V_Y], self[PC_FLOOR] + self[PC_V_FLOOR]

    def at(self) -> Tuple[int, int, int]:
        """
        info: Get the current location of conscious.
        :return: Tuple[int, int, int]
        """
        return self[PC_X], self[PC_Y], self[PC_FLOOR]



"""
Copyright 2021 Charles McMarrow

This script holds a the backrooms API and console interface.
"""

# built-in
import argparse
import cProfile
from typing import Optional, Tuple, Union

# backrooms
from . import backrooms_error
from .backrooms_builtins import get_builtins
from .portal import Portal
from .translator import translator, Handlers, Handler, load_dir
from .whisper import NOTSET, enable_whisper


class BackRoomsError(backrooms_error.BackroomsError):
    pass


def _str_to_bool(boolean: str):
    return boolean.lower() not in ("false", "f", "", "0")


def backrooms() -> None:    # TODO write tests
    """
    info: Console Interface
    :return: None
    """
    try:
        parser = argparse.ArgumentParser(description="backrooms")
        parser.add_argument("file",
                            type=str,
                            action="store",
                            help="path to main file")
        parser.add_argument("--lost-count",
                            default=0,
                            type=int,
                            action="store",
                            help="set lost count")
        parser.add_argument("--lost-rule-count",
                            default=0,
                            type=int,
                            action="store",
                            help="set lost rule count")
        parser.add_argument("--error-on-space",
                            default=False,
                            type=_str_to_bool,
                            action="store",
                            help="errors if portal lands on a space")
        parser.add_argument("--builtins",
                            default=True,
                            type=_str_to_bool,
                            action="store",
                            help="include built-in libraries")
        parser.add_argument("--whisper",
                            default=NOTSET,
                            type=str,
                            action="store",
                            help="set the log level [notset, debug, info, warning, error, critical]")
        parser.add_argument("--profile",
                            default=False,
                            type=_str_to_bool,
                            action="store",
                            help="profiles backrooms")
        args = parser.parse_args()
        br = backrooms_api(code=args.file,
                           lost_count=args.lost_count,
                           lost_rule_count=args.lost_rule_count,
                           error_on_space=args.error_on_space,
                           br_builtins=args.builtins,
                           whisper_level=args.whisper)
        if args.profile:
            with cProfile.Profile() as profiler:
                br()
            print(flush=True)
            profiler.print_stats()
        else:
            br()
    except backrooms_error.BackroomsError as e:
        print(f"ERROR: {e}")
    except KeyboardInterrupt:
        print("Keyboard Interrupt!")


def backrooms_api(code: Union[str, Handler, Handlers],
                  inputs: Optional[Tuple[str, ...]] = None,
                  sys_output: bool = True,
                  catch_output: bool = False,
                  lost_count: int = 0,
                  lost_rule_count: int = 0,
                  error_on_space: bool = False,
                  br_builtins: bool = True,
                  whisper_level: str = NOTSET) -> Portal:
    """
    info: An API to backrooms.
    :param code: Union[str, Handler, Handlers]
        str: Will treat str as main file and load its dir.
        Handler: Will load just the single Handler.
        Handlers: Will load the Handlers.
    :param inputs: Optional[Tuple[str, ...]]
    :param sys_output: bool
    :param catch_output: bool
    :param lost_count: int
    :param lost_rule_count: int
    :param error_on_space: bool
    :param br_builtins: bool
        Only adds builtins if code is str or Handler.
    :param whisper_level: str
    :return: Portal
    """
    try:
        if isinstance(code, str):
            main_handler, handlers = load_dir(code)
            handlers = [handlers]
            if br_builtins:
                handlers.append(get_builtins())
            rooms = translator(Handlers(main_handler, tuple(handlers)))
        elif isinstance(code, Handler):
            handlers = []
            if br_builtins:
                handlers.append(get_builtins())
            rooms = translator(Handlers(code, tuple(handlers)))
        else:
            rooms = translator(code)

        enable_whisper(whisper_level)

        return Portal(rooms,
                      inputs=inputs,
                      sys_output=sys_output,
                      catch_output=catch_output,
                      lost_count=lost_count,
                      lost_rule_count=lost_rule_count,
                      error_on_space=error_on_space)
    except backrooms_error.BackroomsError as e:
        raise BackRoomsError(e)



"""
Copyright 2021 Charles McMarrow

This script holds a simple portal "CPU".
"""


# built-in
from collections import deque
from string import ascii_letters, digits
from typing import Tuple, Optional, Dict, List

# backrooms
from . import backrooms_error
from .conscious import Conscious, ALIVE, ID, HALT
from . import conscious as c
from .rooms import Rooms
from .rules import Rule, RULES, WorkSpace
from . import whisper


VALID_INPUT_CHARACTERS = set(ascii_letters + digits + ",<.>/?;:'\"[{]}\\|`!@#$%^&*()-_=+")


class PortalError(backrooms_error.BackroomsError):
    @classmethod
    def missing_gate(cls):
        return cls("Missing entry point 'GATE'!")

    @classmethod
    def lost_count(cls):
        return cls("Lost count hit! Got lost in the backrooms!")

    @classmethod
    def lost_rule_count(cls):
        return cls("Lost rule count hit! Got lost in the backrooms!")

    @classmethod
    def error_on_space(cls, x: int, y: int, floor: int):
        return cls(f"Error on space at: ({x}, {y}, {floor})")

    @classmethod
    def start_character_collection(cls,
                                   start_character: str):
        return cls(f"{repr(start_character)} is used more then once in a work space!")


class Portal:
    def __init__(self,
                 rooms: Rooms,
                 consciouses: Optional[Tuple[Conscious, ...]] = None,
                 inputs: Optional[Tuple[str, ...]] = None,
                 sys_output: bool = True,
                 catch_output: bool = False,
                 lost_count: int = 0,
                 lost_rule_count: int = 0,
                 error_on_space: bool = False):
        self._done: bool = False
        self._rooms: Rooms = rooms

        if not consciouses:
            hallway_cord = rooms.find_a_hallway("GATE")
            if hallway_cord is None:
                raise PortalError.missing_gate()
            y, floor = hallway_cord
            consciouses = (Conscious(PC_Y=y, PC_FLOOR=floor, ID=0),)

        self._consciouses: deque = deque(consciouses)
        self._lost_count: int = lost_count
        self._lost_rule_count: int = lost_rule_count
        self._sys_output: bool = sys_output
        if inputs is not None:
            inputs = list(inputs[::-1])
        self._inputs: Optional[List[str]] = inputs
        self._catch_output: bool = catch_output
        self._catch_output_steam: List[str] = []
        self._error_on_space: bool = error_on_space

        work_space = WorkSpace()
        rules_obj = [rule(work_space=work_space) for rule in RULES]
        self._rules: Dict[str: Rule] = {}

        for rule in rules_obj:
            if rule.get_start_character() in self._rules:
                raise PortalError.start_character_collection(rule.get_start_character())
            self._rules[rule.get_start_character()] = rule

        used_ids = set(conscious[ID] for conscious in self._consciouses)
        self._next_free_id = max(used_ids) + 1
        self._free_ids: set = set()

        for free_id in range(self._next_free_id):
            if free_id not in used_ids:
                self._free_ids.add(free_id)

        self._rule_step_visuals: List[Tuple[int, int, int]] = []

    def __call__(self):
        for operation_generator in self:
            for step in operation_generator:
                whisper.debug(f"Step: {step}")

    def __iter__(self):
        return self

    def __next__(self):
        if self.is_done():
            raise StopIteration()

        def _run_rule():
            # check if any consciouses remain
            if not len(self._consciouses):
                self._done = True
                whisper.debug("HALT")
            else:
                # get next conscious
                conscious = self._consciouses.popleft()
                # whisper conscious location
                whisper.info(f"{conscious[ID]} {conscious.at()}: {repr(self._rooms.read(*conscious.at()))}")
                whisper.debug(f"{conscious[c.WORKING_REGISTER]} {conscious[c.R0]} {conscious[c.R1]} "
                              f"{conscious[c.R2]} {conscious[c.R3]} {conscious[c.R4]} {conscious[c.R5]} "
                              f"{conscious[c.R6]} {conscious[c.R7]} {conscious[c.R8]} {conscious[c.R9]}")
                whisper.debug(f"{conscious[c.BRANCH]} {conscious[c.WORK_STACK].peak()} "
                              f"{conscious[c.FUNCTION_STACK].peak()}")
                # get rule
                rule = self._rules.get(self._rooms.read(*conscious.at()))
                self._rule_step_visuals.clear()
                self._rule_step_visuals.append(conscious.at())
                if rule is not None:
                    whisper.debug(f"Rule found: {rule.__class__.__name__}")
                    # run operation "rule"
                    for step, _ in enumerate(rule(self, self._rooms, conscious, self._rule_step_visuals)):
                        step += 1
                        if step == self._lost_rule_count:
                            raise PortalError.lost_rule_count()
                        yield step
                else:
                    # whisper that no rule was found for charter
                    whisper.debug(f"No rule found!")
                    if self._error_on_space and self._rooms.read(*conscious.at()) == " ":
                        raise PortalError.error_on_space(*conscious.at())
                    conscious.step()
                # check if conscious is still alive
                if conscious[ALIVE]:
                    # add conscious back to thread queue
                    self._consciouses.append(conscious)
                else:
                    whisper.debug("not ALIVE")
                    # free conscious id
                    self._free_ids.add(conscious[ID])
                    while self._next_free_id - 1 in self._free_ids:
                        self._free_ids.remove(self._next_free_id - 1)
                        self._next_free_id += -1
                    # check if any consciouses remain
                    if not len(self._consciouses):
                        # program is done running
                        self._done = True
                        whisper.debug("HALT")

                whisper.debug(f"Step visuals: {self._rule_step_visuals}")

                # check if conscious raised HALT
                if conscious[HALT] and not self.is_done():
                    # program is done running
                    self._done = True
                    whisper.debug("HALT")

                # check if lost count has been hit
                if self._lost_count > 0:
                    self._lost_count += -1
                    if not self._lost_count:
                        raise PortalError.lost_count()
        return _run_rule()

    def is_done(self) -> bool:
        return self._done

    def get_rooms(self) -> Rooms:
        return self._rooms

    def get_consciouses(self) -> Tuple[Conscious, ...]:
        return tuple(self._consciouses)

    def new_conscious(self) -> Conscious:
        new_conscious = Conscious()
        if self._free_ids:
            # TODO min(set) is Big-O(n)
            free_id = min(self._free_ids)
            self._free_ids.remove(free_id)
            new_conscious[ID] = free_id
        else:
            new_conscious[ID] = self._next_free_id
            self._next_free_id += 1
        self._consciouses.append(new_conscious)
        return new_conscious

    def read_input(self) -> str:    # TODO write test
        """
        info: Gets input from portal
        :return: str
        """
        data = ""
        if self._inputs is None:
            data = input()
        elif self._inputs:
            data = self._inputs.pop()

        valid_data = ""
        for character in data:
            if character in VALID_INPUT_CHARACTERS:
                valid_data += character
        return valid_data

    def write_output(self, output: str) -> None:    # TODO write test for
        """
        info: Writes out to Portal.
        :param output: str
        :return: None
        """
        if self._sys_output:
            print(output, end="", flush=True)

        if self._catch_output:
            self._catch_output_steam.append(output)

    def get_output_stream(self) -> List[str]:   # TODO write test for
        """
        info: Gets the output stream.
            Note anything can be done to the list returned.
        :return: List[str]
        """
        return self._catch_output_steam

    def get_step_visuals(self) -> List[Tuple[int, int, int]]:   # TODO write test for
        """
        info: Gets step visuals for current rule that is executing.
            Note the list should only be read and the list will change when portal takes another step.
        :return: List[Tuple[int, int, int]]
        """
        return self._rule_step_visuals



"""
Copyright 2021 Charles McMarrow
"""


# backrooms
from backrooms.translator import StringHandler

VARIABLES_NAME = "vars"

VARIABLES = """
~GET
~SET
"""


def get_handler() -> StringHandler:
    return StringHandler(VARIABLES_NAME, VARIABLES)



"""
Copyright 2021 Charles McMarrow

This script holds a simple stack data structure.
"""

# built-in
from typing import List


class StackBottom:
    pass


class StackFrame:
    pass


class Stack:
    def __init__(self):
        """
        info: Simple stack with some stack frames
        """
        self._stack: List[object] = [StackBottom]

    def __bool__(self):
        return not self.is_empty()

    def push(self, item: object) -> None:
        """
        info: Push item to stack unless item is stack bottom.
        :param item: object
        :return: object
        """
        if item is not StackBottom:
            self._stack.append(item)

    def pop(self) -> object:
        """
        info: Pop item from stack unless item is stack bottom.
        :return: object
        """
        if self._stack[-1] is StackBottom:
            return StackBottom
        return self._stack.pop()

    def peak(self) -> object:   # TODO test
        """
        info: Peak at item on top of stack
        :return: object
        """
        return self._stack[-1]

    def is_empty(self) -> bool:
        """
        info: Checks if stack is empty.
        :return: bool
        """
        return self._stack[-1] is StackBottom

    def push_frame(self) -> None:
        """
        info: Push stack frame to stack.
        :return: None
        """
        self.push(StackFrame)

    def pop_frame(self) -> None:
        """
        info: Pop till end of stack or till a frame is found.
        :return: None
        """
        while not self.is_empty():
            if self.pop() is StackFrame:
                return



"""
Copyright 2021 Charles McMarrow

This python module "backrooms" is a Esolang.

backrooms was inspired by:
    * backrooms Creepypasta/MEME
    * ASCIIDOTS Esolang
    * CISC Architecture

Backrooms was designed to be:
    * hackable VIA memory overflow attacks, poor error handling, ect.
    * visually pleasing.
    * enjoy able to write small/medium programs.
    * capable to rewrite all of a program at run-time.
"""

# backrooms
from . import backrooms_error
from . import backrooms
from . import rooms
from . import rules
from . import translator
from . import stack
from . import conscious
from . import portal
from . import whisper
from . import backrooms_builtins

AUTHOR = "Charles McMarrow"

MAJOR, MINOR, MAINTENANCE = 0, 2, 0
VERSION = (MAJOR, MINOR, MAINTENANCE)



"""
Copyright 2021 Charles McMarrow

This script holds a simple whisper_level system "log".
"""

# built-in
import logging
from functools import partial
from typing import Callable

WHISPER = logging.getLogger("backrooms")
WHISPER_HANDLER = logging.StreamHandler()
WHISPER_FORMATTER = logging.Formatter('[%(asctime)s][%(thread)d][%(levelname)s] %(message)s')
WHISPER_HANDLER.setFormatter(WHISPER_FORMATTER)
WHISPER.addHandler(WHISPER_HANDLER)
WHISPER_RUNNING = False


NOTSET = "NOTSET"
DEBUG = "DEBUG"
INFO = "INFO"
WARNING = "WARNING"
ERROR = "ERROR"
CRITICAL = "CRITICAL"


WHISPER_LEVEL_STR_TO_INT = {NOTSET: logging.NOTSET,
                            DEBUG: logging.DEBUG,
                            INFO: logging.INFO,
                            WARNING: logging.WARNING,
                            ERROR: logging.ERROR,
                            CRITICAL: logging.CRITICAL}


def enable_whisper(level: str = DEBUG) -> None:
    """
    info: Sets the whisper_level leve.
    :param level: str
    :return: None
    """
    global WHISPER_RUNNING

    level = level.upper()
    if level != NOTSET:
        WHISPER_RUNNING = True
    else:
        WHISPER_RUNNING = False
    level = WHISPER_LEVEL_STR_TO_INT.get(level, logging.NOTSET)
    WHISPER.setLevel(level)


def _whisper(level: Callable, data: str) -> None:
    """
    info: Checks if logging is enabled, then logs data.
        This is done because python logging slows down the program even
        if the logger is not enabled.
    :param level: Callable
    :param data: str
    :return: None
    """
    if WHISPER_RUNNING:
        level(data)
        WHISPER_HANDLER.flush()


debug = partial(_whisper, WHISPER.debug)
info = partial(_whisper, WHISPER.info)
warning = partial(_whisper, WHISPER.warning)
error = partial(_whisper, WHISPER.error)
critical = partial(_whisper, WHISPER.critical)



"""
Copyright 2021 Charles McMarrow
"""

# built-in
from typing import Generator, Tuple, List, Dict, Union, Optional, Callable
import string

# backrooms
import backrooms    # import backrooms to avoid circular imports
from .backrooms_error import BackroomsError
from . import conscious as c
from .rooms import Rooms
from .stack import StackFrame, StackBottom


class RuleError(BackroomsError):
    @classmethod
    def bad_start_character(cls,
                            start_character: str):
        return cls(f"{repr(start_character)} is not a valid start character!")

    @classmethod
    def start_character_collection(cls,
                                   start_character: str):
        return cls(f"{repr(start_character)} is used more then once in a work space!")

    @classmethod
    def bad_hope_start_character(cls, name: str):
        return cls(f"Hope start character most be a single digit not: {repr(name)}!")


def _to_int(obj: Union[int, str, StackFrame, StackBottom, None]) -> Optional[int]:
    if isinstance(obj, str):
        try:
            return int(obj)
        except ValueError:
            return
    elif obj is None:
        return
    elif obj is StackFrame:
        return
    elif obj is StackBottom:
        return
    return obj


def _to_string(obj: Union[int, str, StackFrame, StackBottom, None]) -> str:
    if isinstance(obj, int):
        return str(obj)
    elif obj is None:
        return "None"
    elif obj is StackFrame:
        return "StackFrame"
    elif obj is StackBottom:
        return "StackBottom"
    return obj


SHIFTER = "SHIFTER"


class WorkSpace(dict):
    def __init__(self):
        work_space = {SHIFTER: set()}
        super(WorkSpace, self).__init__(work_space)


class Rule:
    def __init__(self,
                 start_character: str,
                 work_space: WorkSpace):
        if len(start_character) != 1:
            raise RuleError.bad_hope_start_character(start_character)
        self._start_character: str = start_character
        self._work_space: WorkSpace = work_space

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        raise NotImplementedError()

    def get_start_character(self) -> str:
        return self._start_character

    def get_work_space(self) -> WorkSpace:
        return self._work_space


class RuleModule(Rule):
    def __init__(self,
                 start_character: str,
                 work_space: WorkSpace,
                 rules: Tuple):
        super(RuleModule, self).__init__(start_character, work_space)
        rules_obj = [rule(work_space) for rule in rules]
        self._rules: Dict[str, Rule] = {}
        for rule in rules_obj:
            if rule.get_start_character() in self._rules:
                raise RuleError.start_character_collection(rule.get_start_character())
            self._rules[rule.get_start_character()] = rule

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        rule_step_visuals.append(conscious.at())
        yield
        rule = rooms.read(*conscious.at())
        if rule in self._rules:
            for _ in self._rules[rule](portal, rooms, conscious, rule_step_visuals):
                yield


class Halt(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Halt, self).__init__("~", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        yield
        if rooms.read(*conscious.at()) == "h":
            conscious.step()
            rule_step_visuals.append(conscious.at())
            yield
            if rooms.read(*conscious.at()) == "a":
                conscious.step()
                conscious[c.HALT] = True


class Output(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Output, self).__init__("e", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        output = conscious[c.WORK_STACK].peak()
        if output is StackFrame:
            output = "StackFrame"
        elif output is StackBottom:
            output = "StackBottom"
        portal.write_output(output)
        yield


class Input(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Input, self).__init__("c", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        conscious[c.WORK_STACK].push(portal.read_input())
        yield


class Hope(Rule):
    def __init__(self,
                 start_character: str,
                 work_space: WorkSpace):
        if start_character not in string.digits:
            raise RuleError.bad_hope_start_character(start_character)

        super(Hope, self).__init__(start_character, work_space)
        self._jump_count = int(start_character)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        yield
        for _ in range(self._jump_count):
            conscious.step()


class One(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(One, self).__init__("1", work_space)


class Two(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(Two, self).__init__("2", work_space)


class Three(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(Three, self).__init__("3", work_space)


class Four(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(Four, self).__init__("4", work_space)


class Five(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(Five, self).__init__("5", work_space)


class Six(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(Six, self).__init__("6", work_space)


class Seven(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(Seven, self).__init__("7", work_space)


class Eighth(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(Eighth, self).__init__("8", work_space)


class Nine(Hope):
    def __init__(self,
                 work_space: WorkSpace):
        super(Nine, self).__init__("9", work_space)


class Shifter(Rule):
    def __init__(self,
                 vector_x: int,
                 vector_y: int,
                 vector_floor_level: int,
                 start_character: str,
                 work_space: WorkSpace):
        super(Shifter, self).__init__(start_character, work_space)
        self.get_work_space()[SHIFTER].add(self.get_start_character())
        self._vector_x: int = vector_x
        self._vector_y: int = vector_y
        self._vector_floor_level: int = vector_floor_level

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        if conscious[c.BRANCH](conscious):
            conscious[c.PC_V_X] = self._vector_x
            conscious[c.PC_V_Y] = self._vector_y
            conscious[c.PC_V_FLOOR] = self._vector_floor_level
            conscious.step()
            yield
            if rooms.read(*conscious.at()) == self.get_start_character():
                rule_step_visuals.append(conscious.at())
                conscious.step()
                yield
                skip_count = 0
                while True:
                    rule_step_visuals.append(conscious.at())
                    if rooms.read(*conscious.at()) == "!":
                        skip_count += 1
                    elif rooms.read(*conscious.at()) in self.get_work_space()[SHIFTER]:
                        if not skip_count:
                            break
                        skip_count += -1
                    conscious.step()
                    yield
        else:
            conscious.step()
            yield


class Right(Shifter):
    def __init__(self,
                 work_space: WorkSpace):
        super(Right, self).__init__(1, 0, 0, ">", work_space)


class Left(Shifter):
    def __init__(self,
                 work_space: WorkSpace):
        super(Left, self).__init__(-1, 0, 0, "<", work_space)


class Up(Shifter):
    def __init__(self,
                 work_space: WorkSpace):
        super(Up, self).__init__(0, 1, 0, "^", work_space)


class Down(Shifter):
    def __init__(self,
                 work_space: WorkSpace):
        super(Down, self).__init__(0, -1, 0, "v", work_space)


class DownUpper(Shifter):
    def __init__(self,
                 work_space: WorkSpace):
        super(DownUpper, self).__init__(0, -1, 0, "V", work_space)


class Upper(Shifter):
    def __init__(self,
                 work_space: WorkSpace):
        super(Upper, self).__init__(0, 0, 1, "{", work_space)


class Lower(Shifter):
    def __init__(self,
                 work_space: WorkSpace):
        super(Lower, self).__init__(0, 0, -1, "}", work_space)


class Branch(Rule):
    def __init__(self,
                 branch_function: Callable,
                 start_character: str,
                 work_space: WorkSpace):
        super(Branch, self).__init__(start_character, work_space)
        self._branch_function = branch_function

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious[c.BRANCH] = self._branch_function
        conscious.step()
        yield


class Clear(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(Clear, self).__init__(c.BRANCH_CLEAR, "C", work_space)


class LessThanZero(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(LessThanZero, self).__init__(c.BRANCH_LESS_THAN_ZERO, "L", work_space)


class GreaterThanZero(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(GreaterThanZero, self).__init__(c.BRANCH_GREATER_THAN_ZERO, "G", work_space)


class Zero(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(Zero, self).__init__(c.BRANCH_ZERO, "Z", work_space)


class NotZero(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(NotZero, self).__init__(c.BRANCH_NOT_ZERO, "N", work_space)


class IsInteger(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(IsInteger, self).__init__(c.BRANCH_IS_INTEGER, "I", work_space)


class IsString(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(IsString, self).__init__(c.BRANCH_IS_STRING, "S", work_space)


class IsNone(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(IsNone, self).__init__(c.BRANCH_IS_NONE, "O", work_space)


class IsStackFrame(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(IsStackFrame, self).__init__(c.BRANCH_IS_STACK_FRAME, "F", work_space)


class IsStackBottom(Branch):
    def __init__(self,
                 work_space: WorkSpace):
        super(IsStackBottom, self).__init__(c.BRANCH_IS_STACK_BOTTOM, "B", work_space)


class Pop(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Pop, self).__init__("p", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious[c.WORK_STACK].pop()
        conscious.step()
        yield


class Switch(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Switch, self).__init__("z", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item_2 = conscious[c.WORK_STACK].pop()
        item_1 = conscious[c.WORK_STACK].pop()
        conscious[c.WORK_STACK].push(item_2)
        conscious[c.WORK_STACK].push(item_1)
        conscious.step()
        yield


class Duplicate(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Duplicate, self).__init__("d", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        work_space = conscious[c.WORK_STACK]
        work_space.push(work_space.peak())
        conscious.step()
        yield


class Worker(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Worker, self).__init__("q", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        yield
        if rooms.read(*conscious.at()).isdigit():
            conscious[c.WORKING_REGISTER] = "R" + rooms.read(*conscious.at())
            rule_step_visuals.append(conscious.at())
            conscious.step()
            yield


class Keep(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Keep, self).__init__("k", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        conscious[conscious[c.WORKING_REGISTER]] = conscious[c.WORK_STACK].peak()
        yield


class Store(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Store, self).__init__("s", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        conscious[c.WORK_STACK].push(conscious[conscious[c.WORKING_REGISTER]])
        yield


def _read(rooms: Rooms,
          conscious: c.Conscious,
          rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
    type_item = rooms.read(*conscious.at())
    if type_item == "i":
        rule_step_visuals.append(conscious.at())
        conscious.step()
        yield
        if rooms.read(*conscious.at()) in string.digits + "+-":
            new_integer = rooms.read(*conscious.at())
            rule_step_visuals.append(conscious.at())
            conscious.step()
            yield
            while rooms.read(*conscious.at()).isdigit():
                new_integer += rooms.read(*conscious.at())
                rule_step_visuals.append(conscious.at())
                conscious.step()
                yield
            try:
                conscious[c.WORK_STACK].push(int(new_integer))
            except ValueError:
                pass
    elif type_item == "s":
        rule_step_visuals.append(conscious.at())
        conscious.step()
        yield
        start_character = rooms.read(*conscious.at())
        new_string = ""
        rule_step_visuals.append(conscious.at())
        conscious.step()
        yield
        while rooms.read(*conscious.at()) != start_character:
            new_string += rooms.read(*conscious.at())
            rule_step_visuals.append(conscious.at())
            conscious.step()
            yield
        conscious[c.WORK_STACK].push(new_string)
        rule_step_visuals.append(conscious.at())
        conscious.step()
    elif type_item == "n":
        rule_step_visuals.append(conscious.at())
        conscious.step()
        conscious[c.WORK_STACK].push(None)
    elif type_item == "f":
        rule_step_visuals.append(conscious.at())
        conscious.step()
        conscious[c.WORK_STACK].push(StackFrame)


class Read(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Read, self).__init__("r", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        yield
        for _ in _read(rooms, conscious, rule_step_visuals):
            yield


def _write(rooms: Rooms,
           conscious: c.Conscious,
           rule_step_visuals: List[Tuple[int, int, int]]) -> str:
    item = conscious[c.WORK_STACK].pop()
    if item is None:
        return "n"
    elif item is StackBottom:
        return "n"
    elif item is StackFrame:
        return "f"
    elif isinstance(item, int):
        return "i" + str(item)
    rule_step_visuals.append(conscious.at())
    start_character = rooms.read(*conscious.at())
    conscious.step()
    return f"s{start_character}{item}{start_character}"


class Write(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Write, self).__init__("w", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        yield
        raw_data = "r" + _write(rooms, conscious, rule_step_visuals)
        yield
        for character in raw_data:
            rooms.write(*conscious.at(), character=character)
            rule_step_visuals.append(conscious.at())
            conscious.step()
            yield


class StringLength(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringLength, self).__init__("l", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, str):
            conscious[c.WORK_STACK].push(len(item))
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringCast(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringCast, self).__init__("c", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item = conscious[c.WORK_STACK].pop()
        conscious[c.WORK_STACK].push(_to_string(item))
        conscious.step()
        yield


class StringAt(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringAt, self).__init__("a", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        at = conscious[c.WORK_STACK].pop()
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, str) and isinstance(at, int):
            try:
                conscious[c.WORK_STACK].push(item[at])
            except IndexError:
                conscious[c.WORK_STACK].push(None)
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringByte(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringByte, self).__init__("b", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, str) and len(item) == 1:
            conscious[c.WORK_STACK].push(ord(item))
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringSplit(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringSplit, self).__init__("s", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        at = conscious[c.WORK_STACK].pop()
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, str) and isinstance(at, int):
            back = item[at:]
            front = item[:at]
            conscious[c.WORK_STACK].push(front)
            conscious[c.WORK_STACK].push(back)
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringJoin(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringJoin, self).__init__("j", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        back = conscious[c.WORK_STACK].pop()
        front = conscious[c.WORK_STACK].pop()
        if isinstance(front, str) and isinstance(back, str):
            conscious[c.WORK_STACK].push(front + back)
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringEqual(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringEqual, self).__init__("e", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        string_2 = conscious[c.WORK_STACK].pop()
        string_1 = conscious[c.WORK_STACK].pop()
        if isinstance(string_1, str) and isinstance(string_2, str):
            conscious[c.WORK_STACK].push(int(string_1 == string_2))
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringIn(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringIn, self).__init__("i", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        string_2 = conscious[c.WORK_STACK].pop()
        string_1 = conscious[c.WORK_STACK].pop()
        if isinstance(string_1, str) and isinstance(string_2, str):
            conscious[c.WORK_STACK].push(int(string_1 in string_2))
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringUpper(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringUpper, self).__init__("u", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, str):
            conscious[c.WORK_STACK].push(item.upper())
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringLower(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringLower, self).__init__("o", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, str):
            conscious[c.WORK_STACK].push(item.lower())
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class StringModule(RuleModule):
    def __init__(self,
                 work_space: WorkSpace):
        super(StringModule, self).__init__("b",
                                           work_space,
                                           (StringLength,
                                            StringCast,
                                            StringAt,
                                            StringByte,
                                            StringSplit,
                                            StringJoin,
                                            StringEqual,
                                            StringIn,
                                            StringUpper,
                                            StringLower))


class IntegerCast(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerCast, self).__init__("c", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item = conscious[c.WORK_STACK].pop()
        conscious[c.WORK_STACK].push(_to_int(item))
        conscious.step()
        yield


class IntegerOperation(Rule):
    def __init__(self,
                 operation: Callable,
                 start_character: str,
                 work_space: WorkSpace):
        super(IntegerOperation, self).__init__(start_character, work_space)
        self._operation: Callable = operation

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        back = conscious[c.WORK_STACK].pop()
        front = conscious[c.WORK_STACK].pop()
        if isinstance(front, int) and isinstance(back, int):
            try:
                conscious[c.WORK_STACK].push(self._operation(front, back))
            except ZeroDivisionError:
                conscious[c.WORK_STACK].push(None)
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class IntegerAdd(IntegerOperation):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerAdd, self).__init__(int.__add__, "a", work_space)


class IntegerSubtract(IntegerOperation):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerSubtract, self).__init__(int.__sub__, "s", work_space)


class IntegerMultiply(IntegerOperation):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerMultiply, self).__init__(int.__mul__, "m", work_space)


class IntegerDivide(IntegerOperation):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerDivide, self).__init__(int.__floordiv__, "d", work_space)


class IntegerModular(IntegerOperation):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerModular, self).__init__(int.__mod__, "o", work_space)


class IntegerPower(IntegerOperation):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerPower, self).__init__(int.__floordiv__, "p", work_space)


class IntegerByte(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerByte, self).__init__("b", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, int) and 0 <= item < 256:
            conscious[c.WORK_STACK].push(chr(item))
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class IntegerAbsolute(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerAbsolute, self).__init__("l", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, int):
            conscious[c.WORK_STACK].push(abs(item))
        else:
            conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class IntegerModule(RuleModule):
    def __init__(self,
                 work_space: WorkSpace):
        super(IntegerModule, self).__init__("i",
                                            work_space,
                                            (IntegerCast,
                                             IntegerAdd,
                                             IntegerSubtract,
                                             IntegerMultiply,
                                             IntegerDivide,
                                             IntegerModular,
                                             IntegerPower,
                                             IntegerByte,
                                             IntegerAbsolute))


class X(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(X, self).__init__("x", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious[c.WORK_STACK].push(conscious[c.PC_X])
        conscious.step()
        yield


class Y(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Y, self).__init__("y", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious[c.WORK_STACK].push(conscious[c.PC_Y])
        conscious.step()
        yield


class Floor(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(Floor, self).__init__("f", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious[c.WORK_STACK].push(conscious[c.PC_FLOOR])
        conscious.step()
        yield


class ThreadThread(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(ThreadThread, self).__init__("t", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious.step()
        new_conscious = portal.new_conscious()
        new_conscious[c.PC_X] = conscious[c.PC_X]
        new_conscious[c.PC_Y] = conscious[c.PC_Y]
        new_conscious[c.PC_FLOOR] = conscious[c.PC_FLOOR]
        new_conscious[c.PC_V_X] = conscious[c.PC_V_X]
        new_conscious[c.PC_V_Y] = conscious[c.PC_V_Y]
        new_conscious[c.PC_V_FLOOR] = conscious[c.PC_V_FLOOR]
        yield


class ThreadJoin(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(ThreadJoin, self).__init__("j", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        if conscious[c.ID] != 0:
            conscious[c.ALIVE] = False
        else:
            conscious.step()
        yield


class ThreadID(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(ThreadID, self).__init__("i", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        conscious[c.WORK_STACK].push(conscious[c.ID])
        conscious.step()
        yield


class ThreadModule(RuleModule):
    def __init__(self,
                 work_space: WorkSpace):
        super(ThreadModule, self).__init__("t",
                                           work_space,
                                           (ThreadThread,
                                            ThreadJoin,
                                            ThreadID))


class HallwayCall(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayCall, self).__init__("c", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        hallway = conscious[c.WORK_STACK].pop()
        if isinstance(hallway, str):
            hallway = rooms.get_hallway_location(conscious[c.PC_FLOOR], hallway)
        elif isinstance(hallway, int) and rooms.find_hallway_location(hallway, conscious[c.PC_FLOOR]) != hallway:
            hallway = None
        if isinstance(hallway, (int, str)):
            conscious[c.FUNCTION_STACK].push(conscious[c.PC_X])
            conscious[c.FUNCTION_STACK].push(conscious[c.PC_Y])
            conscious[c.FUNCTION_STACK].push(conscious[c.PC_FLOOR])
            conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_X])
            conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_Y])
            conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_FLOOR])
            conscious[c.PC_X] = 0
            conscious[c.PC_Y] = hallway
        else:
            conscious.step()
        yield


class HallwayLevelCall(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayLevelCall, self).__init__("l", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        floor = conscious[c.WORK_STACK].pop()
        hallway = conscious[c.WORK_STACK].pop()
        if isinstance(floor, str):
            floor = rooms.get_floor_level(floor)
        if isinstance(floor, int):
            if isinstance(hallway, str):
                hallway = rooms.get_hallway_location(floor, hallway)
            elif isinstance(hallway, int) and rooms.find_hallway_location(hallway, floor) != hallway:
                hallway = None
            if isinstance(hallway, (int, str)):
                conscious[c.FUNCTION_STACK].push(conscious[c.PC_X])
                conscious[c.FUNCTION_STACK].push(conscious[c.PC_Y])
                conscious[c.FUNCTION_STACK].push(conscious[c.PC_FLOOR])
                conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_X])
                conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_Y])
                conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_FLOOR])
                conscious[c.PC_X] = 0
                conscious[c.PC_Y] = hallway
                conscious[c.PC_FLOOR] = floor
            else:
                conscious.step()
        else:
            conscious.step()
        yield


class HallwayReturn(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayReturn, self).__init__("r", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        v_floor = conscious[c.FUNCTION_STACK].pop()
        v_y = conscious[c.FUNCTION_STACK].pop()
        v_x = conscious[c.FUNCTION_STACK].pop()
        floor = conscious[c.FUNCTION_STACK].pop()
        y = conscious[c.FUNCTION_STACK].pop()
        x = conscious[c.FUNCTION_STACK].pop()
        if isinstance(v_x, int) and isinstance(v_y, int) and isinstance(v_floor, int):
            if isinstance(x, int) and isinstance(y, int) and isinstance(floor, int):
                conscious[c.PC_X] = x
                conscious[c.PC_Y] = y
                conscious[c.PC_FLOOR] = floor
                conscious[c.PC_V_X] = v_x
                conscious[c.PC_V_Y] = v_y
                conscious[c.PC_V_FLOOR] = v_floor
        conscious.step()
        yield


class HallwayGetName(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayGetName, self).__init__("n", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        floor = conscious[c.WORK_STACK].pop()
        hallway = conscious[c.WORK_STACK].pop()
        if isinstance(hallway, int) and isinstance(floor, int):
            hallway = rooms.find_hallway_location(hallway, floor)
            if hallway is not None:
                conscious[c.WORK_STACK].push(rooms.get_hallway_name(hallway, floor))
            else:
                conscious[c.WORK_STACK].push(None)
        conscious.step()
        yield


class HallwayGetLocation(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayGetLocation, self).__init__("g", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        floor = conscious[c.WORK_STACK].pop()
        hallway = conscious[c.WORK_STACK].pop()
        if isinstance(hallway, str) and isinstance(floor, int):
            conscious[c.WORK_STACK].push(rooms.get_hallway_location(floor, hallway))
        elif isinstance(hallway, int) and isinstance(floor, int):
            conscious[c.WORK_STACK].push(rooms.find_hallway_location(hallway, floor))
        conscious.step()
        yield


class HallwaySet(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwaySet, self).__init__("s", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        floor = conscious[c.WORK_STACK].pop()
        hallway = conscious[c.WORK_STACK].pop()
        hallway_name = conscious[c.WORK_STACK].pop()
        if isinstance(floor, int) and isinstance(hallway, int)\
                and (isinstance(hallway_name, str) or hallway_name is None):
            rooms.set_hallway_name(hallway, floor, hallway_name)    # error handling
        conscious.step()
        yield


class HallwayRemove(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayRemove, self).__init__("d", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        hallway = conscious[c.WORK_STACK].pop()
        floor = conscious[c.WORK_STACK].pop()
        if isinstance(hallway, int) and isinstance(floor, int):
            rooms.remove_hallway(hallway, floor)
        conscious.step()
        yield


class HallwayModule(RuleModule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayModule, self).__init__("h",
                                            work_space,
                                            (HallwayCall,
                                             HallwayLevelCall,
                                             HallwayReturn,
                                             HallwayGetLocation,
                                             HallwayGetName,
                                             HallwaySet,
                                             HallwayRemove))


class HallwayGetFloorName(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayGetFloorName, self).__init__("n", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        floor = conscious[c.WORK_STACK].pop()
        if isinstance(floor, int):
            conscious[c.WORK_STACK].push(rooms.get_floor_name(floor))
        conscious.step()
        yield


class HallwayGetFloorLevel(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwayGetFloorLevel, self).__init__("l", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        floor_name = conscious[c.WORK_STACK].pop()
        if isinstance(floor_name, str):
            conscious[c.WORK_STACK].push(rooms.get_floor_level(floor_name))
        conscious.step()
        yield


class HallwaySetFloorLevel(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(HallwaySetFloorLevel, self).__init__("s", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        floor_name = conscious[c.WORK_STACK].pop()
        floor_level = conscious[c.WORK_STACK].pop()
        if isinstance(floor_level, int) and (isinstance(floor_name, str) or floor_name is None):
            rooms.set_floor_name(floor_level, floor_name)
        conscious.step()
        yield


class LevelModule(RuleModule):
    def __init__(self,
                 work_space: WorkSpace):
        super(LevelModule, self).__init__("l",
                                          work_space,
                                          (HallwayGetFloorName,
                                           HallwayGetFloorLevel,
                                           HallwaySetFloorLevel))


class ReadFlip(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(ReadFlip, self).__init__("r", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        x = conscious[c.PC_X]
        y = conscious[c.PC_Y]
        floor = conscious[c.PC_FLOOR]
        v_x = conscious[c.PC_V_X] * -1
        v_y = conscious[c.PC_V_Y] * -1
        v_floor = conscious[c.PC_V_FLOOR] * -1
        conscious.step()
        yield
        for _ in _read(rooms, conscious, rule_step_visuals):
            yield
        conscious[c.PC_X] = x
        conscious[c.PC_Y] = y
        conscious[c.PC_FLOOR] = floor
        conscious[c.PC_V_X] = v_x
        conscious[c.PC_V_Y] = v_y
        conscious[c.PC_V_FLOOR] = v_floor
        conscious.step()
        conscious.step()
        yield


class WriteFlip(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(WriteFlip, self).__init__("w", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        x = conscious[c.PC_X]
        y = conscious[c.PC_Y]
        floor = conscious[c.PC_FLOOR]
        v_x = conscious[c.PC_V_X] * -1
        v_y = conscious[c.PC_V_Y] * -1
        v_floor = conscious[c.PC_V_FLOOR] * -1
        conscious.step()
        yield
        for character in ">1vur" + _write(rooms, conscious, rule_step_visuals):
            rooms.write(*conscious.at(), character=character)
            rule_step_visuals.append(conscious.at())
            conscious.step()
            yield
        conscious[c.PC_X] = x
        conscious[c.PC_Y] = y
        conscious[c.PC_FLOOR] = floor
        conscious[c.PC_V_X] = v_x
        conscious[c.PC_V_Y] = v_y
        conscious[c.PC_V_FLOOR] = v_floor
        conscious.step()
        conscious.step()
        yield


class DynamicDump(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(DynamicDump, self).__init__("d", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        v_floor = conscious[c.WORK_STACK].pop()
        v_y = conscious[c.WORK_STACK].pop()
        v_x = conscious[c.WORK_STACK].pop()
        floor = conscious[c.WORK_STACK].pop()
        y = conscious[c.WORK_STACK].pop()
        x = conscious[c.WORK_STACK].pop()
        item = conscious[c.WORK_STACK].pop()
        if isinstance(item, int):
            item = str(item)
        elif item is None or item is StackBottom:
            item = "n"
        elif item is StackFrame:
            item = "f"
        if isinstance(x, int) and isinstance(y, int) and isinstance(floor, int):
            if isinstance(v_x, int) and isinstance(v_y, int) and isinstance(v_floor, int):
                for character in item:
                    rooms.write(x, y, floor, character)
                    x += v_x
                    y += v_y
                    floor += v_floor
                    yield
        conscious.step()
        yield


class DoubleDuplicate(Rule):
    def __init__(self,
                 work_space: WorkSpace):
        super(DoubleDuplicate, self).__init__("o", work_space)

    def __call__(self,
                 portal: 'backrooms.portal.Portal',
                 rooms: Rooms,
                 conscious: c.Conscious,
                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
        """
        info: Runs a rule.
        :param portal: Portal
        :param rooms: Rooms
        :param conscious: Conscious
        :param rule_step_visuals: List[Tuple[int, int, int]]
        :return: Generator[None, None, None]
        """
        item_2 = conscious[c.WORK_STACK].pop()
        item_1 = conscious[c.WORK_STACK].pop()
        for _ in range(2):
            conscious[c.WORK_STACK].push(item_1)
            conscious[c.WORK_STACK].push(item_2)
        conscious.step()
        yield


class UncommonModule(RuleModule):
    def __init__(self,
                 work_space: WorkSpace):
        super(UncommonModule, self).__init__("u",
                                             work_space,
                                             (ReadFlip,
                                              WriteFlip,
                                              DynamicDump,
                                              DoubleDuplicate))


RULES = (Halt,
         Output,
         Input,
         One,
         Two,
         Three,
         Four,
         Five,
         Six,
         Seven,
         Eighth,
         Nine,
         Right,
         Left,
         Up,
         Down,
         DownUpper,
         Upper,
         Lower,
         Clear,
         LessThanZero,
         GreaterThanZero,
         Zero,
         NotZero,
         IsInteger,
         IsString,
         IsNone,
         IsStackFrame,
         IsStackBottom,
         Pop,
         Duplicate,
         Switch,
         Worker,
         Keep,
         Store,
         Read,
         Write,
         StringModule,
         IntegerModule,
         X,
         Y,
         Floor,
         ThreadModule,
         HallwayModule,
         LevelModule,
         UncommonModule)



"""
Copyright 2021 Charles McMarrow
"""

# built-in
from distutils.core import setup
import os

with open(os.path.join("README.rst")) as readme:
    long_description = readme.read()

setup(
    name="backrooms",
    version="0.2.0",
    author="Charles McMarrow",
    author_email="Charles.McMarrow.4@gmail.com",
    url="https://github.com/cmcmarrow/backrooms",
    license="Apache Software License 2.0",
    description="3D, CISC Architecture and Esolang",
    description_content_type="text/plain",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    packages=["backrooms",
              "backrooms.backrooms_builtins"],

    extras_require={"dev": ["wheel", "check-manifest", "twine", "pyinstaller"]},

    classifiers=["Development Status :: 3 - Alpha",
                 "Environment :: Console",
                 "Intended Audience :: Developers",
                 "Natural Language :: English",
                 "Operating System :: Microsoft :: Windows",
                 "Operating System :: POSIX :: Linux",
                 "Operating System :: MacOS :: MacOS X",
                 "Programming Language :: Python :: 3.8",
                 "Programming Language :: Python :: 3.9"],

    entry_points={"console_scripts": ["backrooms = backrooms.backrooms:backrooms"]}
)



"""
Copyright 2021 Charles McMarrow

This script is the core of backrooms.
It holds the backrooms memory structure and everything need it for it to operate.
"""

# built-in
from functools import lru_cache
from string import ascii_letters, digits
from typing import Optional, Dict, Tuple, List, Set, Union

# backrooms
from .backrooms_error import BackroomsError

CHARACTER_SET = set([chr(character) for character in range(256)])
NAME_SET = set(ascii_letters + digits + "_")


class RoomsError(BackroomsError):
    @classmethod
    def bad_name(cls, bad_name: str) -> 'RoomsError':
        """
        info: Used to indicate a bad name was given.
        :param bad_name: str
        :return: RoomsError
        """
        return cls(f"{repr(bad_name)} is not a name!")

    @classmethod
    def bad_character(cls, bad_character: str) -> 'RoomsError':
        """
        info: Used to indicate a bad character was given.
        :param bad_character:
        :return: RoomsError
        """
        return cls(f"{repr(bad_character)} is not a character!")


def is_character(character: str) -> bool:
    """
    info: Checks if a character is valid.
    :param character: str
    :return: bool
    """
    return character in CHARACTER_SET


@lru_cache(2048)
def is_name(name: str) -> bool:
    """
    info: Checks if a name is valid.
    :param name: str
    :return: bool
    """
    for character in name:
        if character not in NAME_SET:
            return False
    return bool(name)


class Rooms:
    def __init__(self):
        """
        info: 3D memory structure that stores a single ASCII letter in a cell.
        """
        self._floors: Dict[int: Dict[Tuple[int, int]], str] = {}
        self._floor_levels_to_names: Dict[int, str] = {}
        self._floors_names_to_levels: Dict[str, int] = {}

        self._hallways: Dict[int, List[int]] = {}
        self._hallways_set: Dict[int: Set[int]] = {}
        self._hallway_locations_to_names: Dict[int, Dict[int, str]] = {}
        self._hallway_names_to_locations: Dict[int, Dict[str, int]] = {}

    def read(self,
             x: int,
             y: int,
             floor_level: int) -> str:
        """
        info: Reads a memory cell.
        :param x: int
        :param y: int
        :param floor_level: int
        :return: str
        """
        return self._floors.get(floor_level, {}).setdefault((x, y), " ")

    def write(self,
              x: int,
              y: int,
              floor_level: int,
              character: str) -> None:
        """
        info: _Write a character to a memory cell.
        :param x: int
        :param y: int
        :param floor_level: int
        :param character: str
        :exception RoomsError
            raises RoomsError if a bad character is given.
        :return: None
        """

        if not is_character(character):
            raise RoomsError.bad_character(character)

        if character == " ":
            # " " is the default character save space by removing the cell all together
            if (x, y) in self._floors.setdefault(floor_level, {}):
                del self._floors[floor_level][(x, y)]
        else:
            self._floors.setdefault(floor_level, {})[(x, y)] = character

    def write_line(self,
                   x: int,
                   y: int,
                   floor_level: int,
                   characters: str,
                   vector_x: int = 0,
                   vector_y: int = 0,
                   vector_floor_level: int = 0) -> None:
        """
        info: Writes a characters to memory cells.
            The vector variables get added to the coordinates per character.
        :param x: int
        :param y: int
        :param floor_level: int
        :param characters: str
        :param vector_x: int
        :param vector_y: int
        :param vector_floor_level: int
        :exception RoomsError
            raises RoomsError if a bad character is given.
        :return: None
        """
        for character in characters:
            self.write(x, y, floor_level, character)
            x += vector_x
            y += vector_y
            floor_level += vector_floor_level

    def set_floor_name(self,
                       floor_level: int,
                       floor_name: Optional[str] = None) -> None:
        """
        info: Sets a floor name.
        :param floor_level: int
        :param floor_name: Optional[str]
            If the floor name all ready is use it will just be over writen.
            If the floor name is None then that floor name  will be removed.
        :exception RoomsError
            raises RoomsError if a bad name is given.
        :return: None
        """
        if floor_name is None:
            if floor_level in self._floor_levels_to_names:
                # remove floor name
                del self._floors_names_to_levels[self._floor_levels_to_names[floor_level]]
                del self._floor_levels_to_names[floor_level]
        elif is_name(floor_name):
            # check if name is in use
            if floor_name in self._floors_names_to_levels:
                # remove floor name before overwriting it
                self.set_floor_name(self._floors_names_to_levels[floor_name])
            # add floor name
            self._floor_levels_to_names[floor_level] = floor_name
            self._floors_names_to_levels[floor_name] = floor_level
        else:
            # bad name
            raise RoomsError.bad_name(floor_name)

    def get_floor_name(self,
                       floor_level: int) -> Optional[str]:
        """
        info: Gets a floor name if it has one.
        :param floor_level: int
        :return: Optional[str]
        """
        return self._floor_levels_to_names.get(floor_level)

    def get_floor_level(self,
                        floor_name: str) -> Optional[int]:
        """
        info: Gets floor level if floor name is used.
        :param floor_name: str
        :return: Optional[str]
        """
        return self._floors_names_to_levels.get(floor_name)

    def set_hallway_name(self,
                         y: int,
                         floor_level: int,
                         hallway_name: Optional[str] = None) -> None:
        """
        info: Makes a hallway and gives it a name.
        :param y: int
        :param floor_level: int
        :param hallway_name: Optional[str]
        :return: None
        """
        # add hallway
        if hallway_name is None or is_name(hallway_name):
            if hallway_name is not None:
                old_hallway_y = self.get_hallway_location(floor_level, hallway_name)
                # remove hallway that shares same name
                if old_hallway_y is not None:
                    self.remove_hallway(old_hallway_y, floor_level)
            # find location in hallway list to add new hallway
            hallway_block_in_at = self._find_hallway_at(y, floor_level)
            if hallway_block_in_at is None:
                hallway_block_in_at = -1
            # update hallway structures
            #   TODO list.insert is Big-O(n)
            self._hallways.setdefault(floor_level, []).insert(hallway_block_in_at + 1, y)
            self._hallways_set.setdefault(floor_level, set()).add(y)
            if hallway_name is not None:
                self._hallway_locations_to_names.setdefault(floor_level, {})[y] = hallway_name
                self._hallway_names_to_locations.setdefault(floor_level, {})[hallway_name] = y
        else:
            # bad name
            raise RoomsError.bad_name(hallway_name)

    def remove_hallway(self,
                       y: int,
                       floor_level: int) -> None:
        """
        info: Removes a hallway if it exists.
        :param y: int
        :param floor_level: int
        :return: None
        """
        # remove hallway
        if y in self._hallways_set.setdefault(floor_level, set()):
            # find hallway location
            hallway_block_at = self._find_hallway_at(y, floor_level)
            # update hallway structures
            #   TODO del list is Big-O(n)
            del self._hallways[floor_level][hallway_block_at]
            self._hallways_set[floor_level].remove(y)
            hallway_name = self._hallway_locations_to_names[floor_level].get(y)
            if hallway_name is not None:
                del self._hallway_locations_to_names[floor_level][y]
                del self._hallway_names_to_locations[floor_level][hallway_name]

    def get_hallway_name(self,
                         y: int,
                         floor_level: int) -> Optional[str]:
        """
        info: Get hallway name from location..
        :param y: int
        :param floor_level: int
        :return: Optional[str]
        """
        return self._hallway_locations_to_names.setdefault(floor_level, {}).get(y)

    def get_hallway_location(self,
                             floor_level: int,
                             hallway_name: str) -> Optional[int]:
        """
        info: Get hallway location from coordinates.
        :param floor_level: int
        :param hallway_name: str
        :return: Optional[int]
        """
        return self._hallway_names_to_locations.setdefault(floor_level, {}).get(hallway_name)

    def get_next_hallway_location(self,
                                  hallway_location: int,
                                  floor_level: int) -> Optional[int]:
        """
        info: Get coordinates to next hallway.
        :param hallway_location: int
        :param floor_level: int
        :return: Optional[str]
        """
        hallway_at = self._find_hallway_at(hallway_location, floor_level)
        # check that hallway exists
        if hallway_at is not None:
            hallway_at -= 1
            # check that next hallway exists
            if hallway_at != -1:
                return self._hallways[floor_level][hallway_at]

    def get_past_hallway_location(self,
                                  hallway_location: int,
                                  floor_level: int) -> Optional[int]:
        """
        info: Get coordinates to past hallway.
        :param hallway_location: int
        :param floor_level: int
        :return: Optional[str]
        """
        hallway_at = self._find_hallway_at(hallway_location, floor_level)
        # check that hallway exists
        if hallway_at is not None:
            hallway_at += 1
            # check that past hallway exists
            if hallway_at != len(self._hallways[floor_level]):
                return self._hallways[floor_level][hallway_at]

    def find_hallway_location(self,
                              y: int,
                              floor_level: int) -> Optional[int]:
        """
        info: Find what hallway y is in.
        :param y: int
        :param floor_level: int
        :return: Optional[int]
        """
        hallway_at = self._find_hallway_at(y, floor_level)
        if hallway_at is not None:
            return self._hallways[floor_level][hallway_at]

    def _find_hallway_at(self,
                         y: int,
                         floor_level: int) -> Optional[int]:
        """
        info: Finds what hallway the coordinates are in.
        :param y: int
        :param floor_level: int
        :return: Optional[int]
        """
        hallways = self._hallways.setdefault(floor_level, [])

        # check that hallways is not empty
        if not hallways:
            return

        # check that coordinates is not in front of all hallways
        if y > hallways[0]:
            return

        # check that coordinates is not in the last hallway
        if hallways[-1] >= y:
            return len(hallways) - 1

        block_start = 0
        block_end = len(hallways) - 1

        # Big-O(log n)
        # find last hallway that greater than equal to y
        while True:
            # find the middle of block
            at = block_start + (block_end - block_start)//2
            hallway_location = hallways[at]
            # hallway is behind coordinates
            if hallway_location < y:
                # remove back half of block
                block_end = at
            # hallway is in font of coordinates
            elif hallway_location > y:
                # last hallway is not in front of coordinates
                if y > hallways[at + 1]:
                    return at
                # remove first half of block
                block_start = at
            # coordinates is a hallway coordinates
            else:
                return at

    def remove_floor(self,
                     floor_level: int) -> None:
        """
        info: Removes the specified floor and everything associated with it.
        :param floor_level: int
        :return: None
        """
        floor_name = self.get_floor_name(floor_level)
        # remove floor data
        if floor_name is not None:
            del self._floors_names_to_levels[floor_name]

        if floor_level in self._floor_levels_to_names:
            del self._floor_levels_to_names[floor_level]

        if floor_level in self._floors:
            del self._floors[floor_level]

        # remove hallway data
        if floor_level in self._hallways:
            del self._hallways[floor_level]

        if floor_level in self._hallway_names_to_locations:
            del self._hallway_names_to_locations[floor_level]

        if floor_level in self._hallway_locations_to_names:
            del self._hallway_locations_to_names[floor_level]

    def duplicate_floor(self,
                        floor_level_from: int,
                        floor_level_to: int) -> None:
        """
        info: Copies floor onto another floor and everything associated with it.
        :param floor_level_from: int
        :param floor_level_to: int
        :return: None
        """
        # make sure "to" floor is fully removed
        self.remove_floor(floor_level_to)

        # duplicate floor data
        if floor_level_from in self._floors:
            self._floors[floor_level_to] = self._floors[floor_level_from].copy()

        # duplicate hallway data
        if floor_level_from in self._hallways:
            self._hallways[floor_level_to] = self._hallways[floor_level_from].copy()

        if floor_level_from in self._hallway_names_to_locations:
            self._hallway_names_to_locations[floor_level_to] = self._hallway_names_to_locations[floor_level_from].copy()

        if floor_level_from in self._hallway_locations_to_names:
            self._hallway_locations_to_names[floor_level_to] = self._hallway_locations_to_names[floor_level_from].copy()

    def find_a_hallway(self,
                       hallway_name: str) -> Optional[Tuple[int, int]]:
        """
        info: Finds a hallway with that name.
            Works out from 0. EX: 0, 1, -1, 2, -2, 3, ...
        :param hallway_name: str
        :return: Optional[Tuple[int, int]]
        """
        # TODO write tests
        def _key(key: int) -> Union[int, float]:
            """
            info: Makes sorted works out from 0. EX: 0, 1, -1, 2, -2, 3, ...
            :param key: str
            :return: Union[int,float]
            """
            if key < 0:
                return abs(key) - 0.5
            return key

        for floor in sorted(self._hallway_names_to_locations, key=_key):
            if hallway_name in self._hallway_names_to_locations[floor]:
                return self._hallway_names_to_locations[floor][hallway_name], floor
